"""Validation utilities for common data types and formats."""

import re
from urllib.parse import urlparse
from uuid import UUID

__all__ = (
    "is_email",
    "is_vin",
    "is_year",
    "is_uuid",
    "is_valid_http_url",
)


def is_email(email: str) -> bool:
    import re

    return (
        True
        if re.fullmatch(r"\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b", email)
        else False
    )


def is_vin(vin) -> bool:
    vin_pattern = "[A-HJ-NPR-Z0-9]{17}"
    return bool(re.match(vin_pattern, vin))


def is_year(s: str) -> bool:
    if len(s) != 4:
        return False
    try:
        year = int(s)
        if year < 1:
            return False
    except ValueError:
        return False
    return True


def is_uuid(uuid_to_test, version=4) -> bool:
    if isinstance(uuid_to_test, UUID):
        return True

    try:
        uuid_obj = UUID(uuid_to_test, version=version)
    except ValueError:
        return False
    return str(uuid_obj) == uuid_to_test


def is_valid_http_url(url: str) -> bool:
    try:
        result = urlparse(url)
        return result.scheme in ("http", "https") and bool(result.netloc)
    except Exception:
        return False



"""Footprint logging utility for creating structured log entries."""

import logging

__all__ = ("leave",)


def leave(
    log_type: str,
    message: str | None = None,
    **kwargs,
):
    logger = logging.getLogger()

    kwargs["footprint"] = True
    kwargs["message"] = message

    data = dict(
        msg=message,
        extra={
            "details": kwargs,
        },
    )

    error_mapper = {
        "error": logger.error,
        "warning": logger.warning,
        "debug": logger.debug,
    }.get(log_type.lower(), logger.info)

    error_mapper(**data)



"""Database engine/session orchestration utilities.

Centralizes sync/async SQLAlchemy engine creation, session management,
and health checks for both read and write connections.
"""

from __future__ import annotations

from contextlib import asynccontextmanager, contextmanager
from typing import Any, AsyncGenerator, Generator, Optional

from sqlalchemy import NullPool, create_engine, text
from sqlalchemy.ext.asyncio import (AsyncSession, async_sessionmaker,
                                    create_async_engine)
from sqlalchemy.orm import declarative_base, sessionmaker

from ..core.exception import exception_to_dict
from ..log import footprint
from .config import DatabaseConfig

__all__ = ("DatabaseInstance",)


class DatabaseInstance:
    """Manage SQLAlchemy engines and sessions for read/write usage.

    Parameters
    ----------
    config:
        Resolved database configuration holding credentials and connection
        options for read and write connections.
    """

    def __init__(self, config: DatabaseConfig) -> None:
        self.db_backend = (config.get("db_backend") or "postgresql").lower()
        self.db_driver_sync = config.get("db_driver_sync")
        self.db_driver_async = config.get("db_driver_async")
        self.connect_args = config.get("connect_args", {})
        self.db_user = config.get("db_user")
        self.db_password = config.get("db_password")
        self.db_host_write = config.get("db_host")
        self.db_host_read = config.get("db_host_read") or self.db_host_write
        self.db_port = config.get("db_port")
        self.db_name = config.get("db_name")
        self.db_ssl = config.get("db_ssl", False)
        self.db_pool_size = config.get("db_pool_size", None)
        self.db_max_overflow = config.get("db_max_overflow", 0)

        self.active_connections = 0

        # Build connection URLs for sync and async for both write and read.
        db_url = config.get("db_url")

        if db_url:
            self.database_path_write = db_url
            self.async_database_path_write = self._build_async_url(db_url)
        else:
            self.database_path_write = self._build_database_url(
                async_mode=False, host=self.db_host_write
            )
            self.async_database_path_write = self._build_database_url(
                async_mode=True, host=self.db_host_write
            )

        db_url_read = config.get("db_url_read") or db_url
        if db_url_read:
            self.database_path_read = db_url_read
            self.async_database_path_read = self._build_async_url(db_url_read)
        else:
            self.database_path_read = self._build_database_url(
                async_mode=False, host=self.db_host_read
            )
            self.async_database_path_read = self._build_database_url(
                async_mode=True, host=self.db_host_read
            )

        # Database settings
        db_settings = self._initialize_db_settings()

        # Create synchronous engines for write and read.
        self.engine_write = create_engine(self.database_path_write, **db_settings)
        self.engine_read = create_engine(self.database_path_read, **db_settings)

        self.write_session = sessionmaker(
            autocommit=False,
            autoflush=False,
            bind=self.engine_write,
            expire_on_commit=True,
        )
        self.read_session = sessionmaker(
            autocommit=False,
            autoflush=False,
            bind=self.engine_read,
            expire_on_commit=True,
        )

        # Create asynchronous engines for write and read.
        self.async_engine_write = create_async_engine(
            self.async_database_path_write, **db_settings
        )
        self.async_engine_read = create_async_engine(
            self.async_database_path_read, **db_settings
        )
        self.async_write_session = async_sessionmaker(
            bind=self.async_engine_write,
            class_=AsyncSession,
            expire_on_commit=True,
            autocommit=False,
            autoflush=False,
        )
        self.async_read_session = async_sessionmaker(
            bind=self.async_engine_read,
            class_=AsyncSession,
            expire_on_commit=True,
            autocommit=False,
            autoflush=False,
        )

        # Declarative base for ORM models
        self.base = declarative_base(name="Base")

    def _build_async_url(self, db_url: str) -> str:
        if self.db_driver_sync:
            if self.db_driver_async:
                db_url = db_url.replace(
                    f"{self.db_backend}+{self.db_driver_sync}://",
                    f"{self.db_backend}+{self.db_driver_async}://",
                    1,
                )
            else:
                db_url = db_url.replace(
                    f"{self.db_backend}+{self.db_driver_sync}://",
                    f"{self.db_backend}://",
                    1,
                )
        else:
            if self.db_driver_async:
                db_url = db_url.replace(
                    f"{self.db_backend}://",
                    f"{self.db_backend}+{self.db_driver_async}://",
                    1,
                )

        if "postgresql" in db_url:
            return db_url.replace("sslmode=require", "ssl=require", 1)
        else:
            return db_url

    def _build_database_url(
        self, async_mode: bool = False, host: Optional[str] = None
    ) -> str:
        if host is None:
            host = self.db_host_write

        scheme = self.db_backend

        if async_mode and self.db_driver_async:
            scheme += f"+{self.db_driver_async}"
        else:
            if self.db_driver_sync:
                scheme += f"+{self.db_driver_sync}"

        url = f"{scheme}://{self.db_user}:{self.db_password}@{host}:{self.db_port}/{self.db_name}"

        if self.db_ssl:
            if self.db_backend == "postgresql":
                url += "?ssl=require" if async_mode else "?sslmode=require"
            elif self.db_backend == "mysql":
                url += "?ssl-mode=REQUIRED"

        return url

    def _initialize_db_settings(self) -> dict:
        """Build engine keyword arguments based on pooling configuration."""
        settings: dict[str, Any] = {
            "pool_pre_ping": True,
            "echo": False,
            "connect_args": self.connect_args or {},
        }

        if self.db_pool_size:
            settings.update(
                {
                    "pool_size": self.db_pool_size,
                    "pool_recycle": 300,
                    "pool_use_lifo": True,
                    "max_overflow": self.db_max_overflow,
                }
            )
        else:
            settings["poolclass"] = NullPool

        # SSL
        if self.db_ssl and self.db_backend == "mysql":
            settings["connect_args"] = {**settings.get("connect_args", {}), "ssl": {}}

        return settings

    def session_local(self):
        """Create a new synchronous write session."""
        return self.write_session()

    def session_local_read(self):
        """Create a new synchronous read session."""
        return self.read_session()

    def get_db(self, force: Optional[str] = None) -> Generator:
        """Yield a sync session, handling rollback and close on errors.

        Parameters
        ----------
        force:
            If "read" yields a read-only session; otherwise a write session.
        """
        controller = f"{__name__}.get_db"
        if force == "read":
            db = self.session_local_read()
        else:
            db = self.session_local()

        try:
            yield db
        except Exception:
            db.rollback()
            raise
        finally:
            try:
                db.close()
            except Exception as e:
                footprint.leave(
                    log_type="warning",
                    message="We could not close the database because of an error.",
                    controller=controller,
                    subject="Closing Database Connection Error",
                    payload=exception_to_dict(e),
                )

    @contextmanager
    def get_db_cm(self, force: Optional[str] = None):
        """Context manager variant of ``get_db`` for sync sessions."""
        controller = f"{__name__}.get_db_cm"
        if force == "read":
            db = self.session_local_read()
        else:
            db = self.session_local()

        try:
            yield db
        except Exception:
            db.rollback()
            raise
        finally:
            try:
                db.close()
            except Exception as e:
                footprint.leave(
                    log_type="warning",
                    message="We could not close the database because of an error.",
                    controller=controller,
                    subject="Closing Database Connection Error",
                    payload=exception_to_dict(e),
                )

    def get_db_read(self) -> Generator:
        """Yield a synchronous read session."""
        yield from self.get_db(force="read")

    def get_db_write(self) -> Generator:
        """Yield a synchronous write session."""
        yield from self.get_db(force="write")

    @contextmanager
    def get_db_cm_read(self):
        """Context manager for a synchronous read session."""
        with self.get_db_cm(force="read") as db:
            yield db

    @contextmanager
    def get_db_cm_write(self):
        """Context manager for a synchronous write session."""
        with self.get_db_cm(force="write") as db:
            yield db

    def create_tables(self) -> None:
        """Create tables defined on the declarative base in the write
        database."""
        self.base.metadata.create_all(self.engine_write)

    def close_all_connections(self) -> None:
        """Dispose all engines, closing pooled connections."""
        self.engine_write.dispose()
        self.engine_read.dispose()

    def check_database_health(self) -> bool:
        """Ping both write and read databases; return ``True`` if healthy."""
        controller = f"{__name__}.check_database_health"
        try:
            with self.engine_write.connect() as connection:
                connection.execute(text("SELECT 1"))

            with self.engine_read.connect() as connection:
                connection.execute(text("SELECT 1"))
            return True
        except Exception as e:
            footprint.leave(
                log_type="error",
                message="We have issue on database health.",
                controller=controller,
                subject="Database Health Issue",
                payload=exception_to_dict(e),
            )
            return False

    def async_session_local(self) -> AsyncSession:
        """Create a new asynchronous write session."""
        return self.async_write_session()

    def async_session_local_read(self) -> AsyncSession:
        """Create a new asynchronous read session."""
        return self.async_read_session()

    async def async_get_db(
        self, force: Optional[str] = None
    ) -> AsyncGenerator[AsyncSession, None]:
        """Async generator yielding a session with safe cleanup on errors."""
        controller = f"{__name__}.async_get_db"
        if force == "read":
            db = self.async_session_local_read()
        else:
            db = self.async_session_local()

        try:
            yield db
        except Exception:
            await db.rollback()
            raise
        finally:
            try:
                await db.close()
            except Exception as e:
                footprint.leave(
                    log_type="warning",
                    message="We could not close the database because of an error.",
                    controller=controller,
                    subject="Closing Database Connection Error",
                    payload=exception_to_dict(e),
                )

    @asynccontextmanager
    async def async_get_db_cm(
        self, force: Optional[str] = None
    ) -> AsyncGenerator[AsyncSession, None]:
        """Async context manager yielding an async session."""
        if force == "read":
            db = self.async_session_local_read()
        else:
            db = self.async_session_local()

        try:
            yield db
        except Exception:
            await db.rollback()
            raise
        finally:
            await db.close()

    async def async_get_db_read(self) -> AsyncGenerator[AsyncSession, None]:
        """Yield async read sessions from ``async_get_db``."""
        async for db in self.async_get_db(force="read"):
            yield db

    async def async_get_db_write(self) -> AsyncGenerator[AsyncSession, None]:
        """Yield async write sessions from ``async_get_db``."""
        async for db in self.async_get_db(force="write"):
            yield db

    @asynccontextmanager
    async def async_get_db_cm_read(self) -> AsyncGenerator[AsyncSession, None]:
        """Async context manager for a read session."""
        async with self.async_get_db_cm(force="read") as db:
            yield db

    @asynccontextmanager
    async def async_get_db_cm_write(self) -> AsyncGenerator[AsyncSession, None]:
        """Async context manager for a write session."""
        async with self.async_get_db_cm(force="write") as db:
            yield db



from datetime import datetime
from enum import Enum
from typing import Annotated, Any, Dict, List, Literal, Union

from pydantic import BaseModel, ConfigDict, Field, RootModel


class AvailableFilterType(Enum):
    select = "select"
    number = "number"
    date = "date"


class RowsList(RootModel[List[Dict]]):
    pass


class AvailableFilterSelectItem(BaseModel):
    label: str
    value: Any


class AvailableFilterSelect(BaseModel):
    label: str
    name: str
    type: Literal[AvailableFilterType.select]
    value: list[AvailableFilterSelectItem]


class AvailableFilterNumberItem(BaseModel):
    min: int
    max: int


class AvailableFilterNumber(BaseModel):
    label: str
    name: str
    type: Literal[AvailableFilterType.number]
    value: AvailableFilterNumberItem


class AvailableFilterDateItem(BaseModel):
    min: datetime
    max: datetime


class AvailableFilterDate(BaseModel):
    label: str
    name: str
    type: Literal[AvailableFilterType.date]
    value: AvailableFilterDateItem


AvailableFilter = Annotated[
    Union[
        AvailableFilterSelect,
        AvailableFilterNumber,
        AvailableFilterDate,
    ],
    Field(discriminator="type"),
]


class SearchType(Enum):
    select = "select"
    number = "number"
    date = "date"
    search = "search"


class SelectedFilterSelect(BaseModel):
    label: str
    name: str
    type: Literal[SearchType.select]
    value: str | Enum


class SelectedFilterNumberItem(BaseModel):
    min: int
    max: int


class SelectedFilterNumber(BaseModel):
    label: str
    name: str
    type: Literal[SearchType.number]
    value: SelectedFilterNumberItem


class SelectedFilterDateItem(BaseModel):
    min: datetime
    max: datetime


class SelectedFilterDate(BaseModel):
    label: str
    name: str
    type: Literal[SearchType.date]
    value: SelectedFilterDateItem


class SelectedFilterSearch(BaseModel):
    label: str
    name: str
    type: Literal[SearchType.search]
    value: str


SelectedFilter = Annotated[
    Union[
        SelectedFilterSelect,
        SelectedFilterNumber,
        SelectedFilterDate,
        SelectedFilterSearch,
    ],
    Field(discriminator="type"),
]


class PayloadSorting(BaseModel):
    order_by: str | Enum
    sort_by: str | Enum


class Payload(BaseModel):
    search: str = None
    sorting: list[PayloadSorting]
    page: int = 1
    items_per_page: int = 30
    total_row: int
    last_page: int
    has_next: bool

    model_config = ConfigDict(extra="allow")


class SearchResult(BaseModel):
    payload: Payload
    available_filters: List[AvailableFilter] | None = None
    selected_filters: List[SelectedFilter] | None = None
    rows_data: RowsList | None = None



import asyncio
import json
import time
from typing import Dict, Optional

from redis.asyncio import Redis as AsyncRedis

from .common import CommonMethods, REDIS_EXCEPTIONS
from .message import Message
from ..core.exception import exception_to_dict
from ..core.retry import retry_wrapper
from ..log import footprint
from ..redis.connection import RedisInstance


class AsyncRedisStreamer(CommonMethods):
    """
    Asynchronous Redis Streams consumer with:
      - Decoupled fan-out across microservices (one group per service).
      - Bounded at-most-once (per group) via a ZSET de-dup window.
      - Lower network load using adaptive sleeping.
      - Connection pooling for efficient connection reuse.
    """

    def __init__(
        self,
        redis_instance: RedisInstance,
        consumer_name: str,
        dedup_window_ms: int = 7 * 24 * 60 * 60 * 1000,
    ):
        super().__init__(redis_instance, consumer_name, dedup_window_ms)

        self._redis_client: Optional[AsyncRedis] = None
        self._maintenance_task: Optional[asyncio.Task] = None
        self._cleanup_channels_task: Optional[asyncio.Task] = None

    async def _reconnect(self):
        """Attempt to reconnect to Redis with retries asynchronously."""
        controller = f"{__name__}.AsyncRedisStreamer._reconnect"

        for attempt in range(3):
            try:
                footprint.leave(
                    log_type="warning",
                    subject="Attempting to reconnect to Redis",
                    message=f"Attempt {attempt + 1} of 3",
                    controller=controller,
                )
                if self._redis_client is not None:
                    await self._redis_client.close()
                self._redis_client = None
                await self._redis_instance.reset_async_pool()
                client = await self._get_client()
                if await client.ping():
                    footprint.leave(
                        log_type="info",
                        subject="Successfully reconnected to Redis",
                        controller=controller,
                        message=f"Successfully reconnected to Redis (attempt {attempt + 1} of 3)",
                    )
                    return
            except REDIS_EXCEPTIONS as e:
                footprint.leave(
                    log_type="error",
                    subject="Redis reconnection attempt failed",
                    message=f"Attempt {attempt + 1} failed.",
                    controller=controller,
                    payload=exception_to_dict(e),
                )
                await asyncio.sleep(2 ** attempt)  # Exponential backoff

        footprint.leave(
            log_type="error",
            subject="Failed to reconnect to Redis after 3 attempts",
            controller=controller,
            message="Failed to reconnect to Redis",
        )
        raise ConnectionError("Failed to reconnect to Redis after 3 attempts")

    async def _get_client(self) -> AsyncRedis:
        """Lazily initialize and return the async Redis client."""
        if self._redis_client is None:
            self._redis_client = (
                await self._redis_instance.get_async_redis_client()
            )
        return self._redis_client

    @retry_wrapper()
    async def _consumer_group_exists(
        self, channel_name: str, consumer_group: str
    ) -> bool:
        try:
            client = await self._get_client()
            groups = await client.xinfo_groups(channel_name)
            return any(
                group["name"].decode("utf-8") == consumer_group for group in groups
            )
        except REDIS_EXCEPTIONS as e:
            await self._reconnect()
            raise e
        except Exception:
            return False

    @retry_wrapper()
    async def send_message(self, channel: str, message: Message):
        """Send message asynchronously."""
        try:
            footprint.leave(
                log_type="info",
                subject="Sending single message asynchronously",
                controller=f"{__name__}.AsyncRedisStreamer.send_message",
                message=f"Sending message to channel '{channel}' asynchronously.",
                payload={"channel": channel, "message_name": message.name},
            )
            client = await self._get_client()
            await client.xadd(channel, message.get_json_encoded())
        except REDIS_EXCEPTIONS as e:
            await self._reconnect()
            raise e

    @retry_wrapper()
    async def subscribe(self, channel_name: str, start_from_latest: bool = True):
        """Subscribe to a Redis stream channel asynchronously."""
        controller = f"{__name__}.AsyncRedisStreamer.subscribe"
        listener_name = self.listener_name
        group = self._group_name(channel_name, listener_name)

        if not await self._consumer_group_exists(channel_name, group):
            try:
                start_id = "$" if start_from_latest else "0-0"
                client = await self._get_client()
                await client.xgroup_create(channel_name, group, start_id, mkstream=True)
                footprint.leave(
                    log_type="info",
                    subject="Subscription created",
                    message=f"Listener {listener_name} has been subscribed to {channel_name}.",
                    controller=controller,
                    payload={
                        "channel": channel_name,
                        "group": group,
                        "listener": listener_name,
                        "start_from_latest": start_from_latest,
                    },
                )
            except REDIS_EXCEPTIONS as e:
                await self._reconnect()
                raise e
            except Exception as e:
                footprint.leave(
                    log_type="error",
                    subject="Error creating consumer group",
                    message=f"Error creating consumer group {group} for channel {channel_name}.",
                    controller=controller,
                    payload=exception_to_dict(e),
                )

        self._subscriptions.append((channel_name, listener_name, group))
        return self

    @retry_wrapper()
    async def _reserve_once(
        self, processed_key: str, message_id: str, now_ms: int
    ) -> bool:
        try:
            client = await self._get_client()
            added = await client.zadd(processed_key, {message_id: now_ms}, nx=True)
            return added == 1
        except REDIS_EXCEPTIONS as e:
            await self._reconnect()
            raise e
        except Exception as e:
            footprint.leave(
                log_type="error",
                subject="Dedup error",
                controller=f"{__name__}.AsyncRedisStreamer._reserve_once",
                message="ZADD NX failed; skipping message to avoid duplicate processing.",
                payload={
                    "error": exception_to_dict(e),
                    "message_id": message_id,
                    "key": processed_key,
                },
            )
            return False

    @retry_wrapper()
    async def _ack_message(self, channel: str, group: str, message_id: str):
        try:
            client = await self._get_client()
            await client.xack(channel, group, message_id)
        except REDIS_EXCEPTIONS as e:
            await self._reconnect()
            raise e
        except Exception as e:
            footprint.leave(
                log_type="error",
                subject="Ack error",
                controller=f"{__name__}.AsyncRedisStreamer._ack_message",
                message="XACK failed.",
                payload={
                    "error": exception_to_dict(e),
                    "channel": channel,
                    "group": group,
                    "message_id": message_id,
                },
            )

    @retry_wrapper()
    async def _consume_one(
        self,
        channel: str,
        consumer_group: str,
        listener_name: str,
        block_time: float,
        count: int = 32,
    ):
        controller = f"{__name__}.AsyncRedisStreamer._consume_one"

        try:
            client = await self._get_client()
            msgs = await client.xreadgroup(
                groupname=consumer_group,
                consumername=self.consumer_instance_name,
                streams={channel: ">"},
                block=int(block_time * 1000),
                count=count,
            )
            if not msgs:
                return

            _, batch = msgs[0]
            processed_key = self._processed_zset_key(channel, consumer_group)
            now_ms = self._server_now_ms()

            for message_id, fields in batch:
                reserved = await self._reserve_once(
                    processed_key, message_id, now_ms
                )
                if not reserved:
                    await self._ack_message(channel, consumer_group, message_id)
                    continue

                try:
                    raw_name = fields.get(b"name")
                    raw_body = fields.get(b"body")
                    if raw_name is None or raw_body is None:
                        raise ValueError("Missing name or body")
                    name = raw_name.decode("utf-8")
                    body = json.loads(raw_body.decode("utf-8"))
                except Exception as e:
                    self._dead_letter(
                        channel,
                        "decode/schema",
                        message_id,
                        {"listener": listener_name, "error": str(e)},
                    )
                    await self._ack_message(channel, consumer_group, message_id)
                    continue

                for handler in self._handlers.get((channel, listener_name), []):
                    try:
                        footprint.leave(
                            log_type="info",
                            subject="Invoking handler",
                            controller=controller,
                            message=f"Invoking message {name} handler.",
                            payload={
                                "listener": listener_name,
                                "name": name,
                                "body": body,
                            },
                        )

                        import inspect
                        if inspect.iscoroutinefunction(handler):
                            await handler(name=name, payload=body)
                        else:
                            handler(name=name, payload=body)

                        footprint.leave(
                            log_type="info",
                            subject="Handler invoked",
                            controller=controller,
                            message=f"Handler for message {name} invoked successfully.",
                            payload={
                                "listener": listener_name,
                                "name": name,
                            },
                        )
                    except Exception as e:
                        self._dead_letter(
                            channel,
                            "handler",
                            message_id,
                            {
                                "listener": listener_name,
                                "handler": handler.__name__,
                                "error": str(e),
                                "name": name,
                            },
                        )
                        break

                await self._ack_message(channel, consumer_group, message_id)

        except REDIS_EXCEPTIONS as e:
            await self._reconnect()
            raise e
        except Exception as e:
            if "NOGROUP" in str(e):
                try:
                    footprint.leave(
                        log_type="warning",
                        subject=f"Need to resubscribe to channel {channel}",
                        message=f"Trying to resubscribe to {channel}.",
                        controller=controller,
                        payload={
                            "listener_name": listener_name,
                            "channel": channel,
                            "consumer_group": consumer_group,
                            "error": exception_to_dict(e),
                        },
                    )
                    await self.subscribe(channel_name=channel)
                except REDIS_EXCEPTIONS as e_resub:
                    footprint.leave(
                        log_type="error",
                        message=f"Error resubscribing to channel {channel}",
                        controller=controller,
                        subject="Resubscribing Error",
                        payload={
                            "error": exception_to_dict(e_resub),
                            "group": consumer_group,
                            "listener": listener_name,
                        },
                    )
                    await self._reconnect()
                except Exception as e_resub:
                    footprint.leave(
                        log_type="error",
                        message=f"Error resubscribing to channel {channel}",
                        controller=controller,
                        subject="Resubscribing Error",
                        payload={
                            "error": exception_to_dict(e_resub),
                            "group": consumer_group,
                            "listener": listener_name,
                        },
                    )
            else:
                footprint.leave(
                    log_type="error",
                    message=f"Error consuming messages from channel {channel}",
                    controller=controller,
                    subject="Consuming Messages Error",
                    payload={
                        "error": exception_to_dict(e),
                        "group": consumer_group,
                        "listener": listener_name,
                    },
                )

    async def _consume_loop(
        self,
        channel: str,
        listener: str,
        group: str,
        block_time: float,
        count: int,
        rest_time: float,
    ):
        """Dedicated loop per channel."""
        controller = f"{__name__}.AsyncRedisStreamer._consume_loop"

        footprint.leave(
            log_type="info",
            message=f"Launching {channel} async consumer task",
            controller=controller,
            subject="Async Redis consumer",
            payload={
                "channel": channel,
                "listener": listener,
                "group": group,
            },
        )

        idle_backoff = rest_time

        while True:
            try:
                before = time.time()
                await self._consume_one(
                    channel, group, listener, block_time, count
                )
                elapsed = time.time() - before

                if elapsed < block_time:
                    idle_backoff = min(idle_backoff * 2, 2.0)
                else:
                    idle_backoff = rest_time

                await asyncio.sleep(idle_backoff)

            except REDIS_EXCEPTIONS as e:
                footprint.leave(
                    log_type="error",
                    subject="Consumer loop error",
                    controller=controller,
                    message=f"Error in consumer loop for channel {channel}. Restarting loop.",
                    payload={"error": exception_to_dict(e), "channel": channel},
                )
                await self._reconnect()
                await asyncio.sleep(5)
            except Exception as e:
                footprint.leave(
                    log_type="error",
                    subject="Consumer loop error",
                    controller=controller,
                    message=f"Error in consumer loop for channel {channel}. Restarting loop.",
                    payload={"error": exception_to_dict(e), "channel": channel},
                )
                await asyncio.sleep(5)

    async def persist_consume(
        self,
        rest_time: float = 0.1,
        block_time: float = 5.0,
        count: int = 32,
        cleanup_interval: float = 300.0,
    ):
        """Continuously consume messages from all subscribed channels asynchronously."""
        controller = f"{__name__}.AsyncRedisStreamer.persist_consume"

        if self._maintenance_task is None or self._maintenance_task.done():
            self._maintenance_task = asyncio.create_task(
                self._maintenance_loop()
            )

        if self._channel_retention and (
            self._cleanup_channels_task is None
            or self._cleanup_channels_task.done()
        ):
            self._cleanup_channels_task = asyncio.create_task(
                self._cleanup_channels_loop(cleanup_interval)
            )

        tasks = {}
        for channel, listener, group in self._subscriptions:
            task_args = (channel, listener, group, block_time, count, rest_time)
            task = asyncio.create_task(self._consume_loop(*task_args))
            tasks[(channel, listener, group)] = (task, task_args)

        while True:
            await asyncio.sleep(60)
            for key, (task, args) in list(tasks.items()):
                if task.done():
                    try:
                        await task
                    except Exception as e:
                        footprint.leave(
                            log_type="warning",
                            subject="Consumer task died",
                            controller=controller,
                            message=f"Consumer task for channel {key[0]} died. Restarting.",
                            payload={
                                "channel": key[0],
                                "listener": key[1],
                                "error": exception_to_dict(e),
                            },
                        )
                    else:
                        footprint.leave(
                            log_type="warning",
                            subject="Consumer task finished unexpectedly",
                            controller=controller,
                            message=f"Consumer task for channel {key[0]} finished. Restarting.",
                            payload={"channel": key[0], "listener": key[1]},
                        )

                    new_task = asyncio.create_task(self._consume_loop(*args))
                    tasks[key] = (new_task, args)

    @retry_wrapper()
    async def maintain_ledgers(self):
        """Async cleanup for dedup ZSETs."""
        controller = f"{__name__}.AsyncRedisStreamer.maintain_ledgers"
        now_ms = self._server_now_ms()
        cutoff = now_ms - self._dedup_window_ms

        client = await self._get_client()

        for channel_name, _, consumer_group in self._subscriptions:
            try:
                key = self._processed_zset_key(channel_name, consumer_group)
                removed = await client.zremrangebyscore(key, min="-inf", max=f"({cutoff}")
            except REDIS_EXCEPTIONS as e:
                await self._reconnect()
                raise e

            if removed:
                footprint.leave(
                    log_type="info",
                    message=f"Purged {removed} async dedup entries",
                    controller=controller,
                    subject="Async dedup ledger maintenance",
                    payload={"key": key, "removed": removed},
                )

    async def _maintenance_loop(self):
        """Async background task for periodic maintenance."""
        controller = f"{__name__}.AsyncRedisStreamer._maintenance_loop"

        while True:
            try:
                await asyncio.sleep(self._ledger_cleanup_interval / 1000)
                await self.maintain_ledgers()
                self._last_ledger_cleanup = self._server_now_ms()
            except REDIS_EXCEPTIONS as e:
                footprint.leave(
                    log_type="error",
                    subject="Maintenance loop error",
                    controller=controller,
                    message="Error in maintenance loop",
                    payload={"error": exception_to_dict(e)},
                )
                await self._reconnect()
                await asyncio.sleep(60)
            except Exception as e:
                footprint.leave(
                    log_type="error",
                    subject="Maintenance loop error",
                    controller=controller,
                    message="Error in maintenance loop",
                    payload={"error": exception_to_dict(e)},
                )
                await asyncio.sleep(60)

    @retry_wrapper()
    async def cleanup_channels(self):
        """Async clean up messages from specified channels based on a retention period."""
        controller = f"{__name__}.AsyncRedisStreamer.cleanup_channels"
        now_ms = self._server_now_ms()

        client = await self._get_client()
        for channel, retention in self._channel_retention.items():
            if retention is None:
                continue

            try:
                group = self._group_name(channel, self.listener_name)
                processed_key = self._processed_zset_key(channel, group)
                cutoff = now_ms - retention

                removed_count = await client.zremrangebyscore(
                    processed_key, min="-inf", max=f"({cutoff}"
                )

                if removed_count:
                    footprint.leave(
                        log_type="info",
                        message=f"Cleaned up {removed_count} old async dedup entries from channel {channel}",
                        controller=controller,
                        subject="Async channel cleanup",
                        payload={"channel": channel, "removed_count": removed_count},
                    )

            except REDIS_EXCEPTIONS as e:
                await self._reconnect()
                raise e
            except Exception as e:
                footprint.leave(
                    log_type="error",
                    subject="Async channel cleanup error",
                    controller=controller,
                    message=f"Error cleaning up channel {channel}",
                    payload={"error": exception_to_dict(e), "channel": channel},
                )

    async def _cleanup_channels_loop(self, cleanup_interval: float = 300.0):
        """Async background task for periodic channel cleanup."""
        controller = f"{__name__}.AsyncRedisStreamer._cleanup_channels_loop"

        while True:
            try:
                await asyncio.sleep(cleanup_interval)
                await self.cleanup_channels()
            except Exception as e:
                footprint.leave(
                    log_type="error",
                    subject="Async cleanup channels loop error",
                    controller=controller,
                    message="Error in async cleanup channels loop",
                    payload={"error": exception_to_dict(e)},
                )
                await asyncio.sleep(60)

    async def cleanup(self):
        """Async cleanup resources."""
        try:
            await self.stop_consuming()

            footprint.leave(
                log_type="info",
                subject="Cleanup completed",
                controller=f"{__name__}.AsyncRedisStreamer.cleanup",
                message="AsyncRedisStreamer cleanup completed",
            )
        except REDIS_EXCEPTIONS as e:
            footprint.leave(
                log_type="error",
                subject="Cleanup error",
                controller=f"{__name__}.AsyncRedisStreamer.cleanup",
                message="Error during cleanup",
                payload={"error": exception_to_dict(e)},
            )
            await self._reconnect()
        except Exception as e:
            footprint.leave(
                log_type="error",
                subject="Cleanup error",
                controller=f"{__name__}.AsyncRedisStreamer.cleanup",
                message="Error during cleanup",
                payload={"error": exception_to_dict(e)},
            )

    async def stop_consuming(self):
        """Stop all consuming tasks and clean up resources."""
        controller = f"{__name__}.AsyncRedisStreamer.stop_consuming"

        footprint.leave(
            log_type="info",
            subject="Stopping consumer",
            controller=controller,
            message="Stopping all consumer tasks",
        )

        try:
            if self._maintenance_task:
                self._maintenance_task.cancel()
                await self._maintenance_task
        except REDIS_EXCEPTIONS as e:
            footprint.leave(
                log_type="error",
                subject="Error stopping maintenance task",
                controller=f"{__name__}.AsyncRedisStreamer.stop_consuming",
                message="Error stopping maintenance task",
                payload={"error": exception_to_dict(e)},
            )
            await self._reconnect()
        except Exception as e:
            footprint.leave(
                log_type="error",
                subject="Error stopping maintenance task",
                controller=f"{__name__}.AsyncRedisStreamer.stop_consuming",
                message="Error stopping maintenance task",
                payload={"error": exception_to_dict(e)},
            )

        try:
            if self._cleanup_channels_task:
                self._cleanup_channels_task.cancel()
                await self._cleanup_channels_task
        except REDIS_EXCEPTIONS as e:
            footprint.leave(
                log_type="error",
                subject="Error stopping cleanup channels task",
                controller=f"{__name__}.AsyncRedisStreamer.stop_consuming",
                message="Error stopping cleanup channels task",
                payload={"error": exception_to_dict(e)},
            )
            await self._reconnect()
        except Exception as e:
            footprint.leave(
                log_type="error",
                subject="Error stopping cleanup channels task",
                controller=f"{__name__}.AsyncRedisStreamer.stop_consuming",
                message="Error stopping cleanup channels task",
                payload={"error": exception_to_dict(e)},
            )

        try:
            if self._redis_client:
                await self._redis_client.close()
                self._redis_client = None
        except REDIS_EXCEPTIONS as e:
            footprint.leave(
                log_type="error",
                subject="Error closing Redis client",
                controller=f"{__name__}.AsyncRedisStreamer.stop_consuming",
                message="Error closing Redis client",
                payload={"error": exception_to_dict(e)},
            )
            await self._reconnect()
        except Exception as e:
            footprint.leave(
                log_type="error",
                subject="Error closing Redis client",
                controller=f"{__name__}.AsyncRedisStreamer.stop_consuming",
                message="Error closing Redis client",
                payload={"error": exception_to_dict(e)},
            )

    async def get_stats(self) -> Dict:
        """Get current statistics about the streamer."""
        stats = {
            "listener_name": self.listener_name,
            "consumer_instance": self.consumer_instance_name,
            "subscriptions": len(self._subscriptions),
            "channels": [sub[0] for sub in self._subscriptions],
            "dedup_window_ms": self._dedup_window_ms,
            "last_ledger_cleanup": self._last_ledger_cleanup,
        }

        if self._maintenance_task:
            stats["maintenance_running"] = not self._maintenance_task.done()

        return stats

    async def __aenter__(self):
        """Async context manager entry."""
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Async context manager exit with cleanup."""
        await self.cleanup()
        return False



import asyncio
import hashlib
import inspect
import json
import zlib
from functools import wraps
from typing import Any, Callable, Dict, List, Set

from redis import Redis

from ..core.exception import exception_to_dict
from ..core.jsonable_encoder import jsonable_encoder
from ..log import footprint
from .connection import RedisInstance

__all__ = (
    "cache_data",
    "cache_function",
    "cache_wrapper",
)


def cache_data(
    response: Dict, cache_key: str, redis_instance: Redis, expire: int | None = None
):
    controller = f"{__name__}.cache_data"
    try:
        compressed_main_value = zlib.compress(
            json.dumps(jsonable_encoder(response)).encode("utf-8")
        )
        redis_instance.delete(cache_key)
        if expire:
            redis_instance.setex(
                name=cache_key, value=compressed_main_value, time=expire
            )
        else:
            redis_instance.set(name=cache_key, value=compressed_main_value)

    except Exception as exception:
        footprint.leave(
            log_type="error",
            message="We faced an error while we want to cache data.",
            controller=controller,
            subject="Error on caching data",
            payload={
                "expire": expire,
                "cache_key": cache_key,
                "error": exception_to_dict(exception),
            },
        )

    return response


def _cache_key_generator(
    namespace: str,
    args: tuple,
    kwargs: Dict[str, Any],
    skip_cache_keys: Set[str],
):
    kwargs_key = {k: v for k, v in kwargs.items() if k not in skip_cache_keys}
    cache_key = ""
    if namespace:
        cache_key += f"{namespace}:"
    if args:
        args_hash = hashlib.sha256(
            json.dumps(args, default=str).encode("utf-8")
        ).hexdigest()
        cache_key += f"{args_hash}:"
    if kwargs_key:
        kwargs_hash = hashlib.sha256(
            json.dumps(kwargs_key, default=str).encode("utf-8")
        ).hexdigest()
        cache_key += f"{kwargs_hash}:"

    return cache_key.rstrip(":")


def _should_cache(
    cache_only_for: List[Dict[str, Any]] | None, kwargs: Dict[str, Any]
) -> bool:
    if cache_only_for is None:
        return True
    for cond in cache_only_for:
        col = cond.get("kwarg")
        if cond.get("operator") == "in" and kwargs.get(col) in cond.get("value"):
            return True
    return False


def _decode_cached_value(cache_compressed: bytes, controller: str):
    try:
        return json.loads(zlib.decompress(cache_compressed).decode("utf-8"))
    except Exception as exception:
        footprint.leave(
            log_type="error",
            message="Error during decompressing or loading cached data.",
            controller=controller,
            subject="Error on reading cache",
            payload={"error": exception_to_dict(exception)},
        )
        return None


def _encode_result(result: Any) -> bytes:
    return zlib.compress(json.dumps(jsonable_encoder(result)).encode("utf-8"))


def _redis_get_sync(redis_client, key: str, controller: str):
    try:
        return redis_client.get(key)
    except Exception as exception:
        footprint.leave(
            log_type="error",
            message="Error while trying to retrieve data from cache.",
            controller=controller,
            subject="Error on get cached data",
            payload={"redis_key": key, "error": exception_to_dict(exception)},
        )
        return None


async def _redis_get_async(redis_client, key: str, controller: str):
    try:
        return await asyncio.to_thread(redis_client.get, key)
    except Exception as exception:
        footprint.leave(
            log_type="error",
            message="Error while trying to retrieve data from cache.",
            controller=controller,
            subject="Error on get cached data",
            payload={"redis_key": key, "error": exception_to_dict(exception)},
        )
        return None


def _redis_write_sync(
    redis_client, key: str, value: bytes, expire: int | None, controller: str
):
    try:
        redis_client.delete(key)
        if expire:
            redis_client.setex(name=key, time=expire, value=value)
        else:
            redis_client.set(name=key, value=value)
    except Exception as exception:
        footprint.leave(
            log_type="error",
            message="Error occurred while caching the result.",
            controller=controller,
            subject="Error on writing cache",
            payload={"redis_key": key, "error": exception_to_dict(exception)},
        )


async def _redis_write_async(
    redis_client, key: str, value: bytes, expire: int | None, controller: str
):
    try:
        await asyncio.to_thread(redis_client.delete, key)
        if expire:
            # redis-py setex signature: setex(name, time, value)
            await asyncio.to_thread(redis_client.setex, key, expire, value)
        else:
            await asyncio.to_thread(redis_client.set, key, value)
    except Exception as exception:
        footprint.leave(
            log_type="error",
            message="Error occurred while caching the result.",
            controller=controller,
            subject="Error on writing cache",
            payload={"redis_key": key, "error": exception_to_dict(exception)},
        )


def cache_function(
    func: Callable,
    redis: "RedisInstance",
    namespace: str,
    expire: int | None = None,
    cache_only_for: List[Dict[str, Any]] = None,
    skip_cache_keys: Set[str] = None,
    *args,
    **kwargs,
):
    """
    Sync cache wrapper; uses sync Redis I/O.
    """
    controller = f"{__name__}.cache_function"
    if skip_cache_keys is None:
        skip_cache_keys = set()

    if not _should_cache(cache_only_for, kwargs):
        return func(*args, **kwargs)

    cache_key = _cache_key_generator(namespace, args, kwargs, skip_cache_keys)

    with redis.get_redis_client() as redis_instance:
        cache_compressed = _redis_get_sync(redis_instance, cache_key, controller)
        if cache_compressed:
            cached = _decode_cached_value(cache_compressed, controller)
            if cached is not None:
                return cached

        # Miss → compute
        result = func(*args, **kwargs)

        # Write
        _redis_write_sync(
            redis_instance,
            cache_key,
            _encode_result(result),
            expire,
            controller,
        )
        return result


async def acache_function(
    func: Callable,
    redis: "RedisInstance",
    namespace: str,
    expire: int | None = None,
    cache_only_for: List[Dict[str, Any]] = None,
    skip_cache_keys: Set[str] = None,
    *args,
    **kwargs,
):
    """
    Async cache wrapper; Redis remains sync but I/O is offloaded via asyncio.to_thread.
    """
    controller = f"{__name__}.acache_function"
    if skip_cache_keys is None:
        skip_cache_keys = set()

    if not _should_cache(cache_only_for, kwargs):
        if inspect.iscoroutinefunction(func):
            return await func(*args, **kwargs)
        return func(*args, **kwargs)

    cache_key = _cache_key_generator(namespace, args, kwargs, skip_cache_keys)

    with redis.get_redis_client() as redis_instance:
        cache_compressed = await _redis_get_async(redis_instance, cache_key, controller)
        if cache_compressed:
            cached = _decode_cached_value(cache_compressed, controller)
            if cached is not None:
                return cached

        # Miss → compute
        if inspect.iscoroutinefunction(func):
            result = await func(*args, **kwargs)
        else:
            result = func(*args, **kwargs)

        # Write
        await _redis_write_async(
            redis_instance,
            cache_key,
            _encode_result(result),
            expire,
            controller,
        )
        return result


def cache_wrapper(
    redis: "RedisInstance",
    namespace: str,
    expire: int | None = None,
    cache_only_for: List[Dict[str, Any]] = None,
    skip_cache_keys: Set[str] = None,
):
    """
    Unified decorator selecting sync/async path with shared helpers.
    """

    def decorator(func: Callable):
        if inspect.iscoroutinefunction(func):

            @wraps(func)
            async def awrapper(*args, **kwargs):
                return await acache_function(
                    func=func,
                    redis=redis,
                    namespace=namespace,
                    expire=expire,
                    cache_only_for=cache_only_for,
                    skip_cache_keys=skip_cache_keys,
                    *args,
                    **kwargs,
                )

            return awrapper
        else:

            @wraps(func)
            def swrapper(*args, **kwargs):
                return cache_function(
                    func=func,
                    redis=redis,
                    namespace=namespace,
                    expire=expire,
                    cache_only_for=cache_only_for,
                    skip_cache_keys=skip_cache_keys,
                    *args,
                    **kwargs,
                )

            return swrapper

    return decorator



import json
from enum import Enum
from typing import Annotated
from uuid import UUID

from fastapi import Header, HTTPException, status
from pydantic import BaseModel, Field, ValidationError


class UserRole(str, Enum):
    manager = "manager"
    administrator = "administrator"
    super_administrator = "super_administrator"


class PermissionType(str, Enum):
    dealer = "dealer"
    bulk_rule = "bulk_rule"
    inventory = "inventory"
    lead = "lead"
    page = "page"


class UserData(BaseModel):
    """Represents the authenticated user's identity, role, and access permissions."""

    id: UUID | None = Field(default=None, description="Unique identifier of the user.")
    role: UserRole | None = Field(
        default=None,
        description="Role assigned to the user, determining their level of access.",
    )
    permissions: dict[UUID, list[PermissionType]] | None = Field(
        default=None,
        description="Mapping of dealer IDs to the list of permissions the user has for each dealer.",
    )

    def check_accessibility(self, dealer_id: UUID | str) -> bool:
        """
        Check whether the user has access rights for the given dealer.

        Args:
            dealer_id (UUID | str): Dealer identifier (UUID or string).

        Returns:
            bool: True if the user has access, False otherwise.
        """
        # Administrators have universal access
        if self.role in {UserRole.super_administrator, UserRole.administrator}:
            return True

        # Normalize dealer_id to UUID
        uuid_dealer_id = (
            dealer_id if isinstance(dealer_id, UUID) else UUID(str(dealer_id))
        )

        # Check permissions mapping
        return bool(self.permissions and uuid_dealer_id in self.permissions)


def get_user_data(
    user_id: Annotated[
        UUID | None,
        Header(
            alias="user-id",
            description="Unique identifier of the user making the request.",
        ),
    ] = None,
    user_role: Annotated[
        UserRole | None,
        Header(
            alias="user-role",
            description="Role assigned to the user making the request.",
        ),
    ] = None,
    user_permissions: Annotated[
        str | None,
        Header(
            alias="user-permissions",
            description=(
                "JSON-encoded mapping of dealer IDs to the list of "
                "permissions granted to the user."
            ),
        ),
    ] = None,
) -> UserData:
    if user_id is None:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED, detail="Missing user-id header"
        )

    perms = None
    if user_permissions:
        try:
            raw = json.loads(user_permissions)
            perms = {UUID(k): [PermissionType(p) for p in v] for k, v in raw.items()}
        except (json.JSONDecodeError, ValueError, TypeError, ValidationError):
            raise HTTPException(
                status_code=400, detail="Invalid user-permissions header JSON"
            )

    return UserData(id=user_id, role=user_role, permissions=perms)



from sqlalchemy import func, union
from sqlalchemy.orm import Session

__all__ = ("get_filters_value",)


def filters_mapper(
    db: Session, pre_conditions: list, joins: list[dict], filter_item, conditions
):
    columns = filter_item.get("columns")
    columns_type = filter_item.get("type", "select")

    if columns_type in {"select", "select_array"}:
        distinct_names_query = []
        for column in columns:
            query = db.query(
                func.jsonb_array_elements_text(column)
                if filter_item.get("is_json", False)
                else column
            )

            for join_item in joins:
                query = query.join(**join_item)

            distinct_names_query.append(
                query.filter(*pre_conditions, *conditions).distinct()
            )

        values = []
        db_values = list(
            map(lambda x: x[0], db.execute(union(*distinct_names_query)).fetchall())
        )
        if columns_type == "select_array":
            for db_value in db_values:
                if isinstance(db_value, list):
                    values.extend(db_value)
        else:
            if isinstance(db_values, list):
                values.extend(db_values)

        enum = filter_item.get("enum")
        labels = filter_item.get("labels", {})
        items = [
            {
                "label": labels.get(enum(value) if enum else value, value),
                "value": enum(value).name if enum else value,
            }
            for value in values
            if value is not None
        ]
        items = sorted(items, key=lambda x: x.get("label"))
    elif columns_type in ["number", "date"]:
        main_query = db.query(
            func.least(*list(map(func.min, columns))),
            func.greatest(*list(map(func.max, columns))),
        )
        for join_item in joins:
            main_query = main_query.join(**join_item)

        min_value, max_value = main_query.filter(*pre_conditions, *conditions).first()
        items = {"min": min_value, "max": max_value}
    else:
        items = None

    return items


def get_filters_value(
    db: Session,
    pre_conditions: list,
    joins: list[dict],
    filters: list,
    names_conditions: dict,
):
    available_filters = []
    for filter_item in filters:
        filter_name_conditions = names_conditions.get(filter_item["name"])
        if filter_name_conditions is None:
            continue
        filter_mapping_result = filters_mapper(
            db=db,
            pre_conditions=pre_conditions,
            joins=joins,
            filter_item=filter_item,
            conditions=filter_name_conditions,
        )
        filter_label = filter_item.get("label")
        filter_type = filter_item.get("type")
        filter_name = filter_item.get("name")
        if filter_mapping_result is not None:
            if filter_type in {"select", "select_array"}:
                available_filters.append(
                    {
                        "label": filter_label,
                        "name": filter_name,
                        "type": filter_type,
                        "value": filter_mapping_result,
                    }
                )
            elif filter_type in ["number", "date"]:
                available_filters.append(
                    {
                        "label": filter_label,
                        "name": filter_name,
                        "type": filter_type,
                        "value": filter_mapping_result,
                    }
                )
            else:
                available_filters.append(
                    {
                        "label": filter_label,
                        "name": filter_name,
                        "type": filter_type,
                    }
                )

    return available_filters



from contextlib import contextmanager

from kafka import KafkaConsumer, KafkaProducer

from ..core.exception import exception_to_dict
from ..log import footprint
from .config import KafkaConfig


class KafkaInstance:
    """Creates KafkaProducer/Consumer, respecting a single URL or individual
    settings."""

    def __init__(self, config: KafkaConfig) -> None:
        self._config = config
        self._base: dict = self._build_base_config()

    def _build_base_config(self) -> dict:
        # If a full URL is provided, use it directly
        kafka_url = self._config.get("kafka_url")
        if kafka_url:
            return {"bootstrap_servers": [kafka_url]}

        # Otherwise build from individual params
        servers = self._config.get("bootstrap_servers")
        if not servers:
            raise ValueError("Either kafka_url or bootstrap_servers must be configured")
        base = {"bootstrap_servers": servers}

        for key in (
            "security_protocol",
            "sasl_mechanism",
            "sasl_plain_username",
            "sasl_plain_password",
            "client_id",
        ):
            value = self._config.get(key)
            if value:
                base[key] = value

        return base

    def get_producer(self, **kwargs) -> KafkaProducer:
        """Return a KafkaProducer that JSON-encodes messages."""
        controller = f"{__name__}.KafkaInstance.get_producer"
        try:
            return KafkaProducer(
                **self._base,
                value_serializer=lambda v: __import__("json").dumps(v).encode(),
                **kwargs,
            )
        except Exception as e:
            footprint.leave(
                log_type="error",
                message="Error creating KafkaProducer",
                controller=controller,
                subject="Kafka Producer",
                payload={
                    "error": exception_to_dict(e),
                },
            )
            raise

    def get_consumer(self, topics: list[str], **kwargs) -> KafkaConsumer:
        """Return a KafkaConsumer subscribed to given topics."""
        controller = f"{__name__}.KafkaInstance.get_consumer"
        try:
            consumer_config = {
                **self._base,
                "group_id": self._config.get("group_id"),
                "auto_offset_reset": self._config.get("auto_offset_reset", "latest"),
                "enable_auto_commit": self._config.get("enable_auto_commit", True),
                **kwargs,
            }
            consumer = KafkaConsumer(**consumer_config)
            consumer.subscribe(topics)
            return consumer
        except Exception as e:
            footprint.leave(
                log_type="error",
                message="Error creating KafkaConsumer",
                controller=controller,
                subject="Kafka Consumer",
                payload={
                    "error": exception_to_dict(e),
                },
            )
            raise

    @contextmanager
    def producer_context(self, **kwargs):
        prod = self.get_producer(**kwargs)
        try:
            yield prod
        finally:
            prod.flush()
            prod.close()

    @contextmanager
    def consumer_context(self, topics: list[str], **kwargs):
        cons = self.get_consumer(topics, **kwargs)
        try:
            yield cons
        finally:
            cons.close()



import os
import re
import time
import socket
import uuid
from collections import defaultdict
from typing import Dict, DefaultDict, Tuple, List, Callable, Optional

from redis.exceptions import ConnectionError, TimeoutError

from dtpyfw.core.exception import exception_to_dict
from dtpyfw.log import footprint
from dtpyfw.redis.connection import RedisInstance

DEFAULT_EXCEPTIONS = (ConnectionError, TimeoutError, Exception)
REDIS_EXCEPTIONS = (ConnectionError, TimeoutError)


class CommonMethods:

    def __init__(
        self,
        redis_instance: RedisInstance,
        consumer_name: str,
        dedup_window_ms: int = 7 * 24 * 60 * 60 * 1000,
    ):
        self.listener_name: str = self._sanitize(consumer_name, maxlen=128)
        self.consumer_instance_name: str = self._gen_consumer_name()

        self._redis_instance = redis_instance

        self._subscriptions: List[Tuple[str, str, str]] = []
        self._handlers: DefaultDict[Tuple[str, str], List[Callable]] = defaultdict(list)

        # Default dedup window: 1 day
        self._dedup_window_ms: int = dedup_window_ms

        # Maintenance control
        self._last_ledger_cleanup = 0
        self._ledger_cleanup_interval = 300_000  # 5 minutes (ms)
        self._channel_retention: Dict[str, Optional[int]] = {}

    def _gen_consumer_name(self) -> str:
        return self._consumer_name_generator(self.listener_name)

    @staticmethod
    def _sanitize(s: str, maxlen: int) -> str:
        s = re.sub(r"[^a-zA-Z0-9._:-]+", "-", s or "")
        return s[:maxlen]

    @staticmethod
    def _server_now_ms() -> int:
        return int(time.time() * 1000)

    @staticmethod
    def _group_name(channel: str, listener_name: str) -> str:
        return f"{channel}:{listener_name}:cg"

    @staticmethod
    def _processed_zset_key(channel: str, group: str) -> str:
        return f"stream:{channel}:group:{group}:processed"

    @classmethod
    def _consumer_name_generator(cls, listener_name: str) -> str:
        host = os.getenv("POD_NAME") or os.getenv("HOSTNAME") or socket.gethostname()
        pid = os.getpid()
        rnd = uuid.uuid4().hex[:8]
        name = ".".join([listener_name, cls._sanitize(host, maxlen=64), str(pid), rnd])
        return cls._sanitize(name, maxlen=200)

    @staticmethod
    def _dead_letter(channel: str, reason: str, message_id: str, extra: Dict):
        try:
            payload = {
                "reason": reason,
                "channel": channel,
                "message_id": message_id
            }
            if extra:
                payload.update(extra)
            footprint.leave(
                log_type="error",
                subject="Message failed",
                controller=f"{__name__}.AsyncRedisStreamer._dead_letter",
                message=f"Message failure on channel '{channel}' (reason={reason})",
                payload=payload,
            )
        except Exception as e:
            footprint.leave(
                log_type="error",
                subject="Dead-letter logging error",
                controller=f"{__name__}.AsyncRedisStreamer._dead_letter",
                message="Failed to log message failure",
                payload={
                    "error": exception_to_dict(e),
                    "channel": channel,
                    "reason": reason,
                    "message_id": message_id,
                },
            )

    def register_channel(
        self,
        channel_name: str,
        retention_ms: int = 24 * 60 * 60 * 1000,
    ):
        """Register channel metadata with optional retention configuration."""
        controller = f"{__name__}.AsyncRedisStreamer.register_channel"

        footprint.leave(
            log_type="info",
            subject="Channel registered",
            message=f"Channel {channel_name} registered.",
            controller=controller,
            payload={
                "channel_name": channel_name,
                "retention_ms": retention_ms,
            },
        )
        self._channel_retention[channel_name] = retention_ms
        return self

    def register_handler(
        self,
        channel_name: str,
        handler_func: Callable,
        listener_name: Optional[str] = None,
    ):
        """Register a handler function for a specific channel."""
        listener = listener_name or self.listener_name
        self._handlers[(channel_name, listener)].append(handler_func)
        footprint.leave(
            log_type="info",
            subject="Handler registered",
            controller=f"{__name__}.AsyncRedisStreamer.register_handler",
            message=f"Handler '{handler_func.__name__}' registered for channel '{channel_name}'.",
            payload={
                "channel_name": channel_name,
                "handler_name": handler_func.__name__,
                "listener_name": listener,
            },
        )
        return self



import json
import threading
import time
from typing import Dict, Any

from redis import Redis

from .common import CommonMethods, REDIS_EXCEPTIONS
from .message import Message
from ..core.exception import exception_to_dict
from ..core.retry import retry_wrapper
from ..log import footprint
from ..redis.connection import RedisInstance


class RedisStreamer(CommonMethods):
    """
    Synchronous Redis Streams consumer with:
      - Decoupled fan-out across microservices (one group per service).
      - Bounded at-most-once (per group) via a ZSET de-dup window.
      - Lower network load using adaptive sleeping.
      - Connection pooling for efficient connection reuse.
    """

    def __init__(
        self,
        redis_instance: RedisInstance,
        consumer_name: str,
        dedup_window_ms: int = 7 * 24 * 60 * 60 * 1000,
    ):
        super().__init__(redis_instance, consumer_name, dedup_window_ms)

        self._redis_client: Redis = self._redis_instance.get_redis_client()
        self._maintenance_thread_started = False

    def _reconnect(self):
        """Attempt to reconnect to Redis with retries."""
        controller = f"{__name__}.RedisStreamer._reconnect"

        for attempt in range(3):
            try:
                footprint.leave(
                    log_type="warning",
                    subject="Attempting to reconnect to Redis",
                    message=f"Attempt {attempt + 1} of 3",
                    controller=controller,
                )
                if self._redis_client:
                    self._redis_client.close()
                self._redis_instance.reset_sync_pool()
                self._redis_client = self._redis_instance.get_redis_client()
                if self._redis_client.ping():
                    footprint.leave(
                        log_type="info",
                        subject="Successfully reconnected to Redis",
                        controller=controller,
                        message=f"Successfully reconnected to Redis (attempt {attempt + 1} of 3)",
                    )
                    return
            except REDIS_EXCEPTIONS as e:
                footprint.leave(
                    log_type="error",
                    subject="Redis reconnection attempt failed",
                    message=f"Attempt {attempt + 1} failed.",
                    controller=controller,
                    payload=exception_to_dict(e),
                )
                time.sleep(2 ** attempt)  # Exponential backoff

        footprint.leave(
            log_type="error",
            subject="Failed to reconnect to Redis after 3 attempts",
            controller=controller,
            message="Failed to reconnect to Redis",
        )
        raise ConnectionError("Failed to reconnect to Redis after 3 attempts")

    @retry_wrapper()
    def _consumer_group_exists(self, channel_name: str, consumer_group: str) -> bool:
        try:
            groups = self._redis_client.xinfo_groups(channel_name)
            return any(
                group["name"].decode("utf-8") == consumer_group for group in groups
            )
        except REDIS_EXCEPTIONS as e:
            self._reconnect()
            raise e
        except Exception as e:
            return False

    @retry_wrapper()
    def send_message(self, channel: str, message: Message):
        """Send single message."""
        try:
            footprint.leave(
                log_type="info",
                subject="Sending single message",
                controller=f"{__name__}.RedisStreamer.send_message",
                message=f"Sending message to channel '{channel}'.",
                payload={"channel": channel, "message_name": message.name},
            )
            fields: Dict[Any, Any] = message.get_json_encoded()
            self._redis_client.xadd(channel, fields)
        except REDIS_EXCEPTIONS as e:
            self._reconnect()
            raise e

    @retry_wrapper()
    def subscribe(self, channel_name: str, start_from_latest: bool = True):
        """Subscribe to a Redis stream channel."""
        controller = f"{__name__}.RedisStreamer.subscribe"
        listener_name = self.listener_name
        group = self._group_name(channel_name, listener_name)

        if not self._consumer_group_exists(channel_name, group):
            try:
                start_id = "$" if start_from_latest else "0-0"
                self._redis_client.xgroup_create(
                    channel_name, group, start_id, mkstream=True
                )
                footprint.leave(
                    log_type="info",
                    subject="Subscription created",
                    message=f"Listener {listener_name} has been subscribed to {channel_name}.",
                    controller=controller,
                    payload={
                        "channel": channel_name,
                        "group": group,
                        "listener": listener_name,
                        "start_from_latest": start_from_latest,
                    },
                )
            except REDIS_EXCEPTIONS as e:
                self._reconnect()
                raise e
            except Exception as e:
                footprint.leave(
                    log_type="error",
                    subject="Error creating consumer group",
                    message=f"Error creating consumer group {group} for channel {channel_name}.",
                    controller=controller,
                    payload=exception_to_dict(e),
                )

        self._subscriptions.append((channel_name, listener_name, group))
        return self

    @retry_wrapper()
    def _reserve_once(self, processed_key: str, message_id: str, now_ms: int) -> bool:
        try:
            added = self._redis_client.zadd(
                processed_key, {message_id: now_ms}, nx=True
            )
            return added == 1
        except REDIS_EXCEPTIONS as e:
            self._reconnect()
            raise e
        except Exception as e:
            footprint.leave(
                log_type="error",
                subject="Dedup error",
                controller=f"{__name__}.RedisStreamer._reserve_once",
                message="ZADD NX failed; skipping message to avoid duplicate processing.",
                payload={
                    "error": exception_to_dict(e),
                    "message_id": message_id,
                    "key": processed_key,
                },
            )
            return False

    @retry_wrapper()
    def _ack_message(self, channel: str, group: str, message_id: str):
        try:
            self._redis_client.xack(channel, group, message_id)
        except REDIS_EXCEPTIONS as e:
            self._reconnect()
            raise e
        except Exception as e:
            footprint.leave(
                log_type="error",
                subject="Ack error",
                controller=f"{__name__}.RedisStreamer._ack_message",
                message="XACK failed.",
                payload={
                    "error": exception_to_dict(e),
                    "channel": channel,
                    "group": group,
                    "message_id": message_id,
                },
            )

    @retry_wrapper()
    def _consume_one(
        self,
        channel: str,
        consumer_group: str,
        listener_name: str,
        block_time: float,
        count: int = 32,
    ):
        controller = f"{__name__}.RedisStreamer._consume_one"
        try:
            msgs = self._redis_client.xreadgroup(
                groupname=consumer_group,
                consumername=self.consumer_instance_name,
                streams={channel: ">"},
                block=int(block_time * 1000),
                count=count,
            )
            if not msgs:
                return

            _, batch = msgs[0]
            processed_key = self._processed_zset_key(channel, consumer_group)
            now_ms = self._server_now_ms()

            for message_id, fields in batch:
                reserved = self._reserve_once(processed_key, message_id, now_ms)
                if not reserved:
                    self._ack_message(channel, consumer_group, message_id)
                    continue

                try:
                    raw_name = fields.get(b"name")
                    raw_body = fields.get(b"body")
                    if raw_name is None or raw_body is None:
                        raise ValueError("Missing required fields 'name' or 'body'.")
                    name = (
                        raw_name.decode("utf-8")
                        if isinstance(raw_name, bytes)
                        else raw_name
                    )
                    body = (
                        json.loads(raw_body.decode("utf-8"))
                        if isinstance(raw_body, (bytes, bytearray))
                        else raw_body
                    )
                except Exception as e:
                    self._dead_letter(
                        channel,
                        "decode/schema",
                        message_id,
                        {"listener": listener_name, "error": str(e)},
                    )
                    self._ack_message(channel, consumer_group, message_id)
                    continue

                for handler in self._handlers.get((channel, listener_name), []):
                    try:
                        footprint.leave(
                            log_type="info",
                            subject="Invoking handler",
                            controller=controller,
                            message=f"Invoking message {name} handler.",
                            payload={
                                "listener": listener_name,
                                "name": name,
                                "body": body,
                            },
                        )
                        handler(name=name, payload=body)
                        footprint.leave(
                            log_type="info",
                            subject="Handler invoked",
                            controller=controller,
                            message=f"Handler for message {name} invoked successfully.",
                            payload={
                                "listener": listener_name,
                                "name": name,
                            },
                        )
                    except Exception as e:
                        self._dead_letter(
                            channel,
                            "handler",
                            message_id,
                            {
                                "listener": listener_name,
                                "handler": handler.__name__,
                                "error": str(e),
                                "name": name,
                            },
                        )
                        break

                self._ack_message(channel, consumer_group, message_id)

        except REDIS_EXCEPTIONS as e:
            self._reconnect()
            raise e
        except Exception as e:
            if "NOGROUP" in str(e):
                try:
                    footprint.leave(
                        log_type="warning",
                        subject=f"Need to resubscribe to channel {channel}",
                        message=f"Trying to resubscribe to {channel}.",
                        controller=controller,
                        payload={
                            "listener_name": listener_name,
                            "channel": channel,
                            "consumer_group": consumer_group,
                            "error": exception_to_dict(e),
                        },
                    )
                    self.subscribe(channel_name=channel)
                except REDIS_EXCEPTIONS as e_resub:
                    footprint.leave(
                        log_type="error",
                        message=f"Error resubscribing to channel {channel}",
                        controller=controller,
                        subject="Resubscribing Error",
                        payload={
                            "error": exception_to_dict(e_resub),
                            "group": consumer_group,
                            "listener": listener_name,
                        },
                    )
                    self._reconnect()
                except Exception as e_resub:
                    footprint.leave(
                        log_type="error",
                        message=f"Error resubscribing to channel {channel}",
                        controller=controller,
                        subject="Resubscribing Error",
                        payload={
                            "error": exception_to_dict(e_resub),
                            "group": consumer_group,
                            "listener": listener_name,
                        },
                    )
            else:
                footprint.leave(
                    log_type="error",
                    message=f"Error consuming messages from channel {channel}",
                    controller=controller,
                    subject="Consuming Messages Error",
                    payload={
                        "error": exception_to_dict(e),
                        "group": consumer_group,
                        "listener": listener_name,
                    },
                )

    def _consume_loop(
        self,
        channel: str,
        listener: str,
        group: str,
        block_time: float,
        count: int,
        rest_time: float,
    ):
        """Dedicated loop per channel."""
        controller = f"{__name__}.RedisStreamer._consume_loop"

        footprint.leave(
            log_type="info",
            message=f"Launching {channel} consumer thread",
            controller=controller,
            subject="Multi-threaded Redis consumer",
            payload={
                "channel": channel,
                "listener": listener,
                "group": group,
            },
        )

        idle_backoff = rest_time

        while True:
            try:
                before = time.time()
                self._consume_one(channel, group, listener, block_time, count)
                elapsed = time.time() - before

                if elapsed < block_time:
                    idle_backoff = min(idle_backoff * 2, 2.0)
                else:
                    idle_backoff = rest_time

                time.sleep(idle_backoff)
            except REDIS_EXCEPTIONS as e:
                footprint.leave(
                    log_type="warning",
                    subject="Consumer loop error",
                    controller=controller,
                    message=f"Error in consumer loop for channel {channel}. Restarting loop.",
                    payload={"error": exception_to_dict(e), "channel": channel},
                )
                self._reconnect()
                time.sleep(5)
            except Exception as e:
                footprint.leave(
                    log_type="warning",
                    subject="Consumer loop error",
                    controller=controller,
                    message=f"Error in consumer loop for channel {channel}. Restarting loop.",
                    payload={"error": exception_to_dict(e), "channel": channel},
                )
                time.sleep(5)

    def persist_consume(
        self, rest_time: float = 0.1, block_time: float = 5.0, count: int = 32
    ):
        """Continuously consume messages from all subscribed channels."""
        controller = f"{__name__}.RedisStreamer.persist_consume"

        if not self._maintenance_thread_started:
            mt = threading.Thread(target=self._maintenance_loop, daemon=True)
            mt.start()
            self._maintenance_thread_started = True

        threads = {}
        for channel, listener, group in self._subscriptions:
            thread_args = (channel, listener, group, block_time, count, rest_time)
            t = threading.Thread(
                target=self._consume_loop,
                args=thread_args,
                daemon=True,
            )
            t.start()
            threads[(channel, listener, group)] = (t, thread_args)

        while True:
            time.sleep(60)
            for key, (thread, args) in list(threads.items()):
                if not thread.is_alive():
                    footprint.leave(
                        log_type="warning",
                        subject="Consumer thread died",
                        controller=controller,
                        message=f"Consumer thread for channel {key[0]} died. Restarting.",
                        payload={"channel": key[0], "listener": key[1]},
                    )
                    new_thread = threading.Thread(
                        target=self._consume_loop, args=args, daemon=True
                    )
                    new_thread.start()
                    threads[key] = (new_thread, args)

    @retry_wrapper()
    def maintain_ledgers(self):
        """Cleanup for dedup ZSETs."""
        controller = f"{__name__}.RedisStreamer.maintain_ledgers"
        now_ms = self._server_now_ms()
        cutoff = now_ms - self._dedup_window_ms

        for channel_name, _, consumer_group in self._subscriptions:
            try:
                key = self._processed_zset_key(channel_name, consumer_group)
                removed = self._redis_client.zremrangebyscore(
                    key, min="-inf", max=f"({cutoff}"
                )
            except REDIS_EXCEPTIONS as e:
                self._reconnect()
                raise e
            if removed:
                footprint.leave(
                    log_type="info",
                    message=f"Purged {removed} dedup entries",
                    controller=controller,
                    subject="Dedup ledger maintenance",
                    payload={"key": key, "removed": removed},
                )

    def _maintenance_loop(self):
        """Background thread for periodic maintenance tasks."""
        controller = f"{__name__}.RedisStreamer._maintenance_loop"

        while True:
            try:
                time.sleep(self._ledger_cleanup_interval / 1000)
                self.maintain_ledgers()
                self._last_ledger_cleanup = self._server_now_ms()
            except REDIS_EXCEPTIONS as e:
                footprint.leave(
                    log_type="error",
                    subject="Maintenance loop error",
                    controller=controller,
                    message="Error in maintenance loop",
                    payload={"error": exception_to_dict(e)},
                )
                self._reconnect()
                time.sleep(60)
            except Exception as e:
                footprint.leave(
                    log_type="error",
                    subject="Maintenance loop error",
                    controller=controller,
                    message="Error in maintenance loop",
                    payload={"error": exception_to_dict(e)},
                )
                time.sleep(60)

    def cleanup(self):
        """Cleanup resources."""
        try:
            footprint.leave(
                log_type="info",
                subject="Cleanup completed",
                controller=f"{__name__}.RedisStreamer.cleanup",
                message="RedisStreamer cleanup completed",
            )
        except Exception as e:
            footprint.leave(
                log_type="error",
                subject="Cleanup error",
                controller=f"{__name__}.RedisStreamer.cleanup",
                message="Error during cleanup",
                payload={"error": exception_to_dict(e)},
            )

    def get_stats(self) -> Dict:
        """Get current statistics about the streamer."""
        stats = {
            "listener_name": self.listener_name,
            "consumer_instance": self.consumer_instance_name,
            "subscriptions": len(self._subscriptions),
            "channels": [sub[0] for sub in self._subscriptions],
            "dedup_window_ms": self._dedup_window_ms,
            "last_ledger_cleanup": self._last_ledger_cleanup,
        }

        return stats

    def __enter__(self):
        """Context manager entry."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context manager exit with cleanup."""
        self.cleanup()
        return False



import json
from dataclasses import dataclass
from typing import Dict


@dataclass
class Message:
    name: str
    body: Dict

    def get_json_encoded(self) -> Dict[str, str]:
        """Return a JSON-encoded representation of the message.
        
        Returns:
            Dict[str, str]: Dictionary with 'name' and JSON-encoded 'body'.
        """
        return {"name": self.name, "body": json.dumps(self.body, default=str)}



"""High-level FastAPI application wrapper and configuration helpers."""

from typing import Any, Dict, Optional, Sequence, Tuple, Union

from fastapi import FastAPI
from starlette.requests import Request

from .middlewares import http_exception, runtime, timer, validation_exception
from .middlewares.user_agent import InternalUserAgentRestriction
from .routes.router import Router

__all__ = ("Application",)


class Application:
    """Wrapper for configuring a FastAPI app using clean OOP structure."""

    def __init__(
        self,
        title: str,
        version: str = "*",
        redoc_url: Optional[str] = "/",
        docs_url: Optional[str] = "/swagger",
        applications: Optional[Sequence[Tuple[str, "Application"]]] = None,
        routers: Optional[
            Union[Sequence[Tuple[str, Sequence[Router]]], Sequence[Router]]
        ] = None,
        gzip_min_size: Optional[int] = None,
        session_middleware_settings: Optional[Dict[str, Any]] = None,
        middlewares: Optional[Sequence[Any]] = None,
        only_internal_user_agent: Optional[bool] = False,
        lifespan: Optional[Any] = None,
        cors_settings: Optional[Dict[str, Any]] = None,
        hide_error_messages: bool = True,
    ) -> None:
        """Initialize the application wrapper."""
        # Core metadata
        self.title = title
        self.version = version
        self.redoc_url = redoc_url
        self.docs_url = docs_url
        self.lifespan = lifespan

        # Sub-apps and routers
        self.applications = applications or ()
        self.routers = routers or ()

        # Middleware configuration
        base_mws = list(middlewares or [])
        if only_internal_user_agent:
            base_mws.insert(0, InternalUserAgentRestriction())

        self.middlewares = (
            timer.Timer(),
            runtime.Runtime(hide_error_messages),
            *base_mws,
        )

        self.gzip_min_size = gzip_min_size
        self.session_settings = session_middleware_settings or {}

        # CORS defaults and overrides
        self.cors_settings: Dict[str, Any] = {
            "allow_origins": ["*"],
            "allow_credentials": True,
            "allow_methods": ["*"],
            "allow_headers": ["*"],
            "expose_headers": [],
            "allow_origin_regex": None,
            "max_age": 600,
        }
        if cors_settings:
            self.cors_settings.update(cors_settings)

        # Build the FastAPI instance
        self.app = self._create_app()

    def _create_app(self) -> FastAPI:
        """Create and configure the internal :class:`FastAPI` instance."""
        app = FastAPI(
            title=self.title,
            version=self.version,
            redoc_url=self.redoc_url,
            docs_url=self.docs_url,
            swagger_ui_parameters={"defaultModelsExpandDepth": -1},
            lifespan=self.lifespan,
        )
        self._configure_gzip(app)
        self._configure_middlewares(app)
        self._configure_session(app)
        self._configure_cors(app)
        self._configure_exception_handlers(app)
        self._mount_sub_applications(app)
        self._include_routers(app)
        return app

    def _configure_gzip(self, app: FastAPI) -> None:
        """Configure gzip compression for the app if enabled."""
        if self.gzip_min_size is not None:
            from fastapi.middleware.gzip import GZipMiddleware

            app.add_middleware(GZipMiddleware, minimum_size=self.gzip_min_size)

    def _configure_session(self, app: FastAPI) -> None:
        """Attach the session middleware when settings are provided."""
        if self.session_settings:
            from starlette.middleware.sessions import SessionMiddleware

            app.add_middleware(SessionMiddleware, **self.session_settings)

    def _configure_middlewares(self, app: FastAPI) -> None:
        """Register custom middlewares with the app."""
        if self.middlewares:
            from starlette.middleware.base import BaseHTTPMiddleware

            for mw in self.middlewares:
                app.add_middleware(BaseHTTPMiddleware, dispatch=mw)

    def _configure_cors(self, app: FastAPI) -> None:
        """Apply CORS settings to the app."""
        if self.cors_settings:
            from fastapi.middleware.cors import CORSMiddleware

            app.add_middleware(CORSMiddleware, **self.cors_settings)

    def _configure_exception_handlers(self, app: FastAPI) -> None:
        """Register common exception handlers with the app."""
        from starlette.exceptions import \
            HTTPException as StarletteHTTPException

        @app.exception_handler(StarletteHTTPException)
        async def _http_exception_handler(
            request: Request, exc: StarletteHTTPException
        ) -> Any:
            """Handle HTTP exceptions using the shared utility."""
            return await http_exception.http_exception_handler(request, exc)

        from fastapi.exceptions import RequestValidationError

        @app.exception_handler(RequestValidationError)
        async def _validation_exception_handler(
            request: Request, exc: RequestValidationError
        ) -> Any:
            """Handle request validation exceptions."""
            return await validation_exception.validation_exception_handler(request, exc)

    def _mount_sub_applications(self, app: FastAPI) -> None:
        """Mount nested applications under the configured prefixes."""
        for prefix, sub_app in self.applications:
            app.mount(prefix, sub_app.get_app())

    def _include_routers(self, app: FastAPI) -> None:
        """Include routers under their prefixes if specified."""
        for item in self.routers:
            if isinstance(item, tuple):
                prefix, routers = item
                for router in routers:
                    app.include_router(
                        router=router.get_router(),
                        prefix=prefix,
                    )
            else:
                app.include_router(
                    router=item.get_router(),
                )

    def get_app(self) -> FastAPI:
        """Return the fully configured FastAPI application."""
        return self.app
